@@ -126,6 +126,7 @@ class BaseVisitor(ast.NodeVisitor):
     self._stack = []
     self._indent = ''
     self._indent_diff = ''
+    self._default_indent_diff = '  '
 
   def visit(self, node):
     self._stack.append(node)
@@ -144,19 +145,23 @@ class BaseVisitor(ast.NodeVisitor):
     self.attr(node, 'suffix', [_ws], default=default)
 
   def indented(self, node, children_attr):
+    children = getattr(node, children_attr)
     prev_indent = self._indent
     prev_indent_diff = self._indent_diff
-    new_diff = fmt.get(node, 'indent')
+    new_diff = fmt.get(children[0], 'indent_diff')
     if new_diff is None:
-      new_diff = '  '
+      new_diff = self._default_indent_diff
     self._indent_diff = new_diff
     self._indent = prev_indent + self._indent_diff
-    for child in getattr(node, children_attr):
+    for child in children:
       yield child
     self.attr(node, 'block_suffix_%s' % children_attr, [])
     self._indent = prev_indent
     self._indent_diff = prev_indent_diff
 
+  def set_default_indent_diff(self, indent):
+    self._default_indent_diff = indent
+
   @contextlib.contextmanager
   def scope(self, node, attr=None, trailing_comma=False, default_parens=False):
     """Context manager to handle a parenthesized scope.
@@ -281,7 +286,11 @@ class BaseVisitor(ast.NodeVisitor):
 
   @block_statement
   def visit_For(self, node):
-    self.attr(node, 'for_keyword', ['for', self.ws], default='for ')
+    if hasattr(ast, 'AsyncFor') and isinstance(node, ast.AsyncFor):
+      self.attr(node, 'for_keyword', ['async', self.ws, 'for', self.ws],
+                default='async for ')
+    else:
+      self.attr(node, 'for_keyword', ['for', self.ws], default='for ')
     self.visit(node.target)
     self.attr(node, 'for_in', [self.ws, 'in', self.ws], default=' in ')
     self.visit(node.iter)
@@ -297,6 +306,9 @@ class BaseVisitor(ast.NodeVisitor):
       for stmt in self.indented(node, 'orelse'):
         self.visit(stmt)
 
+  def visit_AsyncFor(self, node):
+    return self.visit_For(node)
+
   @block_statement
   def visit_With(self, node):
     if hasattr(node, 'items'):
@@ -317,6 +329,9 @@ class BaseVisitor(ast.NodeVisitor):
     for stmt in self.indented(node, 'body'):
       self.visit(stmt)
 
+  def visit_AsyncWith(self, node):
+    return self.visit_With(node)
+
   @abc.abstractmethod
   def check_is_continued_try(self, node):
     pass
@@ -341,7 +356,11 @@ class BaseVisitor(ast.NodeVisitor):
     """
 
   def visit_With_3(self, node):
-    self.attr(node, 'with', ['with', self.ws], default='with ')
+    if hasattr(ast, 'AsyncWith') and isinstance(node, ast.AsyncWith):
+      self.attr(node, 'with', ['async', self.ws, 'with', self.ws],
+                default='async with ')
+    else:
+      self.attr(node, 'with', ['with', self.ws], default='with ')
 
     for i, withitem in enumerate(node.items):
       self.visit(withitem)
@@ -368,13 +387,20 @@ class BaseVisitor(ast.NodeVisitor):
                 default='\n' + self._indent)
     self.attr(node, 'class_def', ['class', self.ws, node.name, self.ws],
               default='class %s' % node.name, deps=('name',))
-    with self.scope(node, 'bases', trailing_comma=bool(node.bases),
+    class_args = getattr(node, 'bases', []) + getattr(node, 'keywords', [])
+    with self.scope(node, 'bases', trailing_comma=bool(class_args),
                     default_parens=True):
       for i, base in enumerate(node.bases):
         self.visit(base)
         self.attr(node, 'base_suffix_%d' % i, [self.ws])
-        if base != node.bases[-1]:
+        if base != class_args[-1]:
           self.attr(node, 'base_sep_%d' % i, [',', self.ws], default=', ')
+      if hasattr(node, 'keywords'):
+        for i, keyword in enumerate(node.keywords):
+          self.visit(keyword)
+          self.attr(node, 'keyword_suffix_%d' % i, [self.ws])
+          if keyword != node.keywords[-1]:
+            self.attr(node, 'keyword_sep_%d' % i, [',', self.ws], default=', ')
     self.attr(node, 'open_block', [self.ws, ':', self.ws_oneline],
               default=':\n')
     for stmt in self.indented(node, 'body'):
@@ -388,9 +414,15 @@ class BaseVisitor(ast.NodeVisitor):
       self.visit(decorator)
       self.attr(node, 'decorator_suffix_%d' % i, [self.ws_oneline],
                 default='\n' + self._indent)
-    self.attr(node, 'function_def',
-              [self.ws, 'def', self.ws, node.name, self.ws],
-              deps=('name',), default='def %s' % node.name)
+    if (hasattr(ast, 'AsyncFunctionDef') and
+        isinstance(node, ast.AsyncFunctionDef)):
+      self.attr(node, 'function_def',
+                [self.ws, 'async', self.ws, 'def', self.ws, node.name, self.ws],
+                deps=('name',), default='async def %s' % node.name)
+    else:
+      self.attr(node, 'function_def',
+                [self.ws, 'def', self.ws, node.name, self.ws],
+                deps=('name',), default='def %s' % node.name)
     # In Python 3, there can be extra args in kwonlyargs
     kwonlyargs = getattr(node.args, 'kwonlyargs', [])
     args_count = sum((len(node.args.args + kwonlyargs),
@@ -410,6 +442,9 @@ class BaseVisitor(ast.NodeVisitor):
     for stmt in self.indented(node, 'body'):
       self.visit(stmt)
 
+  def visit_AsyncFunctionDef(self, node):
+    return self.visit_FunctionDef(node)
+
   @block_statement
   def visit_TryFinally(self, node):
     # Try with except and finally is a TryFinally with the first statement as a
@@ -419,8 +454,10 @@ class BaseVisitor(ast.NodeVisitor):
     # TODO(soupytwist): Find a cleaner solution for differentiating this.
     if len(node.body) == 1 and self.check_is_continued_try(node.body[0]):
       node.body[0].is_continued = True
-    for stmt in node.body:
-      self.visit(stmt)
+      self.visit(node.body[0])
+    else:
+      for stmt in self.indented(node, 'body'):
+        self.visit(stmt)
     self.attr(node, 'open_finally',
               [self.ws, 'finally', self.ws, ':', self.ws_oneline],
               default='finally:\n')
@@ -432,7 +469,7 @@ class BaseVisitor(ast.NodeVisitor):
     if not getattr(node, 'is_continued', False):
       self.attr(node, 'open_try', ['try', self.ws, ':', self.ws_oneline],
                 default='try:\n')
-    for stmt in node.body:
+    for stmt in self.indented(node, 'body'):
       self.visit(stmt)
     for handler in node.handlers:
       self.visit(handler)
@@ -448,7 +485,7 @@ class BaseVisitor(ast.NodeVisitor):
     # Python 3
     self.attr(node, 'open_try', [self.ws, 'try', self.ws, ':', self.ws_oneline],
               default='try:\n')
-    for stmt in node.body:
+    for stmt in self.indented(node, 'body'):
       self.visit(stmt)
     for handler in node.handlers:
       self.visit(handler)
@@ -547,6 +584,11 @@ class BaseVisitor(ast.NodeVisitor):
       self.attr(node, 'equal', [self.ws, '=', self.ws], default=' = ')
       self.visit(node.value)
 
+  @expression
+  def visit_Await(self, node):
+    self.attr(node, 'await', ['await', self.ws], default='await ')
+    self.visit(node.value)
+
   @statement
   def visit_Break(self, node):
     self.token('break')
@@ -628,6 +670,11 @@ class BaseVisitor(ast.NodeVisitor):
         if alias is not node.names[-1]:
           self.attr(node, 'alias_sep_%d' % i, [self.ws, ','], default=',')
 
+  @expression
+  def visit_NamedExpr(self, node):
+    self.visit(target)
+    self.attr(node, 'equal' % i, [self.ws, ':=', self.ws], default=' := ')
+    self.visit(node.value)
 
   @statement
   def visit_Nonlocal(self, node):
@@ -668,13 +715,19 @@ class BaseVisitor(ast.NodeVisitor):
       self.attr(node, 'return_value_prefix', [self.ws], default=' ')
       self.visit(node.value)
 
-  @statement
+  @expression
   def visit_Yield(self, node):
     self.token('yield')
     if node.value:
       self.attr(node, 'yield_value_prefix', [self.ws], default=' ')
       self.visit(node.value)
 
+  @expression
+  def visit_YieldFrom(self, node):
+    self.attr(node, 'yield_from', ['yield', self.ws, 'from', self.ws],
+              default='yield from ')
+    self.visit(node.value)
+
   # ============================================================================
   # == EXPRESSIONS: Anything that evaluates and can be in parens              ==
   # ============================================================================
@@ -819,8 +872,7 @@ class BaseVisitor(ast.NodeVisitor):
     self.visit(node.key)
     self.attr(node, 'key_val_sep', [self.ws, ':', self.ws], default=': ')
     self.visit(node.value)
-    for i, comp in enumerate(node.generators):
-      self.attr(node, 'for_%d' % i, [self.ws, 'for', self.ws], default=' for ')
+    for comp in node.generators:
       self.visit(comp)
     self.attr(node, 'close_dict', [self.ws, '}'], default='}')
 
@@ -865,7 +917,6 @@ class BaseVisitor(ast.NodeVisitor):
       self.attr(node, 'compexp_open', [open_brace, self.ws], default=open_brace)
     self.visit(node.elt)
     for i, comp in enumerate(node.generators):
-      self.attr(node, 'for_%d' % i, [self.ws, 'for', self.ws], default=' for ')
       self.visit(comp)
     if close_brace:
       self.attr(node, 'compexp_close', [self.ws, close_brace],
@@ -1048,11 +1099,13 @@ class BaseVisitor(ast.NodeVisitor):
     assert len(kwonlyargs) == len(kw_defaults)
 
     total_args = sum((len(node.args + kwonlyargs),
+                      len(getattr(node, 'posonlyargs', [])),
                       1 if node.vararg else 0,
                       1 if node.kwarg else 0))
     arg_i = 0
 
-    positional = node.args[:-len(node.defaults)] if node.defaults else node.args
+    pos_args = getattr(node, 'posonlyargs', []) + node.args
+    positional = pos_args[:-len(node.defaults)] if node.defaults else pos_args
     keyword = node.args[-len(node.defaults):] if node.defaults else node.args
 
     for arg in positional:
@@ -1061,6 +1114,9 @@ class BaseVisitor(ast.NodeVisitor):
       if arg_i < total_args:
         self.attr(node, 'comma_%d' % arg_i, [self.ws, ',', self.ws],
                   default=', ')
+      if arg_i == len(getattr(node, 'posonlyargs', [])):
+        self.attr(node, 'posonly_sep', [self.ws, '/', self.ws, ',', self.ws],
+                  default='/, ')
 
     for i, (arg, default) in enumerate(zip(keyword, node.defaults)):
       self.visit(arg)
@@ -1085,8 +1141,7 @@ class BaseVisitor(ast.NodeVisitor):
     elif kwonlyargs:
       # If no vararg, but we have kwonlyargs, insert a naked *, which will
       # definitely not be the last arg.
-      self.token('*')
-      self.token(',')
+      self.attr(node, 'kwonly_sep', [self.ws, '*', self.ws, ',', self.ws]);
 
     for i, (arg, default) in enumerate(zip(kwonlyargs, kw_defaults)):
       self.visit(arg)
@@ -1109,6 +1164,11 @@ class BaseVisitor(ast.NodeVisitor):
 
   @space_around
   def visit_comprehension(self, node):
+    if getattr(node, 'is_async', False):
+      self.attr(node, 'for', [self.ws, 'async', self.ws, 'for', self.ws],
+                default=' async for ')
+    else:
+      self.attr(node, 'for', [self.ws, 'for', self.ws], default=' for ')
     self.visit(node.target)
     self.attr(node, 'in', [self.ws, 'in', self.ws], default=' in ')
     self.visit(node.iter)
@@ -1196,7 +1256,7 @@ class AstAnnotator(BaseVisitor):
       raise AnnotationError(e)
 
   def indented(self, node, children_attr):
-    """Generator white annotates child nodes with their indentation level."""
+    """Generator which annotates child nodes with their indentation level."""
     children = getattr(node, children_attr)
     cur_loc = self.tokens._loc
     next_loc = self.tokens.peek_non_whitespace().start
@@ -1217,15 +1277,15 @@ class AstAnnotator(BaseVisitor):
     indent_token = self.tokens.peek_conditional(
         lambda t: t.type == token_generator.TOKENS.INDENT)
     new_indent = indent_token.src
-    if (not new_indent.startswith(prev_indent) or
-        len(new_indent) <= len(prev_indent)):
-      raise AnnotationError(
-          'Indent detection failed (line %d); inner indentation level is not '
-          'more than the outer indentation.' % cur_loc[0])
+    new_diff = _get_indent_diff(prev_indent, new_indent)
+    if not new_diff:
+      new_diff = ' ' * 4  # Sensible default
+      print('Indent detection failed (line %d); inner indentation level is not '
+            'more than the outer indentation.' % cur_loc[0], file=sys.stderr)
 
     # Set the indent level to the child's indent and iterate over the children
     self._indent = new_indent
-    self._indent_diff = new_indent[len(prev_indent):]
+    self._indent_diff = new_diff
     for child in children:
       yield child
     # Store the suffix at this indentation level, which could be many lines
@@ -1431,3 +1491,53 @@ class AstAnnotator(BaseVisitor):
     else:
       self.tokens.next()
       return token.src + self.ws()
+
+
+def _get_indent_width(indent):
+  width = 0
+  for c in indent:
+    if c == ' ':
+      width += 1
+    elif c == '\t':
+      width += 8 - (width % 8)
+  return width
+
+
+def _ltrim_indent(indent, remove_width):
+  width = 0
+  for i, c in enumerate(indent):
+    if width == remove_width:
+      break
+    if c == ' ':
+      width += 1
+    elif c == '\t':
+      if width + 8 - (width % 8) <= remove_width:
+        width += 8 - (width % 8)
+      else:
+        return ' ' * (width + 8 - remove_width) + indent[i + 1:]
+  return indent[i:]
+
+
+def _get_indent_diff(outer, inner):
+  """Computes the whitespace added to an indented block.
+
+  Finds the portion of an indent prefix that is added onto the outer indent. In
+  most cases, the inner indent starts with the outer indent, but this is not
+  necessarily true. For example, the outer block could be indented to four
+  spaces and its body indented with one tab (effectively 8 spaces).
+
+  Arguments:
+    outer: (string) Indentation of the outer block.
+    inner: (string) Indentation of the inner block.
+  Returns:
+    The string whitespace which is added to the indentation level when moving
+    from outer to inner.
+  """
+  outer_w = _get_indent_width(outer)
+  inner_w = _get_indent_width(inner)
+  diff_w = inner_w - outer_w
+
+  if diff_w <= 0:
+    return None
+
+  return _ltrim_indent(inner, inner_w - diff_w)
